@@ -1,9 +1,11 @@
 #include "media_player.h"
+
 #include <glib.h>
 #ifdef WITH_TRANSCODING
 #include <mysql.h>
 #include <mysql/errmsg.h>
 #endif
+
 #include "obj.h"
 #include "log.h"
 #include "timerthread.h"
@@ -20,29 +22,44 @@
 #include "fix_frame_channel_layout.h"
 #endif
 #include "kernel.h"
-
-
+#include "bufferpool.h"
+#include "uring.h"
 
 #define DEFAULT_AVIO_BUFSIZE 4096
 
-
+typedef enum {
+	MPC_OK = 0,
+	MPC_ERR = -1,
+	MPC_CACHED = 1,
+} mp_cached_code;
 
 #ifdef WITH_TRANSCODING
 static struct timerthread media_player_thread;
 static __thread MYSQL *mysql_conn;
 
 
+struct media_player_cache_packet;
+static void cache_packet_free(struct media_player_cache_packet *p);
+TYPED_GPTRARRAY_FULL(cache_packet_arr, struct media_player_cache_packet, cache_packet_free)
+
+
 struct media_player_cache_index {
 	struct media_player_content_index index;
-	struct rtp_payload_type dst_pt;
+	rtp_payload_type dst_pt;
 };
+TYPED_DIRECT_FUNCS(media_player_direct_hash, media_player_direct_eq, struct media_player)
+TYPED_GHASHTABLE(media_player_ht, struct media_player, struct media_player, media_player_direct_hash,
+		media_player_direct_eq, NULL, NULL) // XXX ref counting players
 struct media_player_cache_entry {
-	bool finished;
+	volatile bool finished;
 	// "unfinished" elements, only used while decoding is active:
 	mutex_t lock;
 	cond_t cond; // to wait for more data to be decoded
 
-	GPtrArray *packets; // read-only except for decoder thread, which uses finished flags and locks
+	cache_packet_arr *packets; // read-only except for decoder thread, which uses finished flags and locks
+	unsigned long duration; // cumulative in ms, summed up while decoding
+	unsigned int kernel_idx; // -1 if not in use
+	media_player_ht wait_queue; // players waiting on decoder to finish
 
 	struct codec_scheduler csch;
 	struct media_player_coder coder; // de/encoder data
@@ -53,7 +70,7 @@ struct media_player_cache_packet {
 	char *buf;
 	str s;
 	long long pts;
-	long long duration;
+	long long duration; // us
 	long long duration_ts;
 };
 
@@ -113,15 +130,28 @@ static void media_player_shutdown(struct media_player *mp) {
 		mp->ssrc_out->parent->seq_diff -= num;
 	}
 
+	if (mp->opts.block_egress)
+		MEDIA_CLEAR(mp->media, BLOCK_EGRESS);
+
 	mp->media = NULL;
 	media_player_coder_shutdown(&mp->coder);
 
+	if (mp->kernel_idx != -1)
+		kernel_stop_stream_player(mp->kernel_idx);
+	else if (mp->cache_entry) {
+		mutex_lock(&mp->cache_entry->lock);
+		if (t_hash_table_is_set(mp->cache_entry->wait_queue))
+			t_hash_table_remove(mp->cache_entry->wait_queue, mp);
+		mutex_unlock(&mp->cache_entry->lock);
+	}
+
 	mp->cache_index.type = MP_OTHER;
 	if (mp->cache_index.file.s)
 		g_free(mp->cache_index.file.s);
 	mp->cache_index.file = STR_NULL;// coverity[missing_lock : FALSE]
 	mp->cache_entry = NULL; // coverity[missing_lock : FALSE]
 	mp->cache_read_idx = 0;
+	mp->kernel_idx = -1;
 }
 #endif
 
@@ -154,8 +184,11 @@ static void __media_player_free(void *p) {
 
 
 // call->master_lock held in W
-struct media_player *media_player_new(struct call_monologue *ml) {
+void media_player_new(struct media_player **mpp, struct call_monologue *ml) {
 #ifdef WITH_TRANSCODING
+	if (*mpp)
+		return;
+
 	//ilog(LOG_DEBUG, "creating media_player");
 
 	uint32_t ssrc = 0;
@@ -164,10 +197,11 @@ struct media_player *media_player_new(struct call_monologue *ml) {
 	struct ssrc_ctx *ssrc_ctx = get_ssrc_ctx(ssrc, ml->ssrc_hash, SSRC_DIR_OUTPUT, ml);
 	ssrc_ctx->next_rtcp = rtpe_now;
 
-	struct media_player *mp = obj_alloc0("media_player", sizeof(*mp), __media_player_free);
+	struct media_player *mp = *mpp = obj_alloc0("media_player", sizeof(*mp), __media_player_free);
 
 	mp->tt_obj.tt = &media_player_thread;
 	mutex_init(&mp->lock);
+	mp->kernel_idx = -1;
 
 	mp->run_func = media_player_read_packet; // default
 	mp->call = obj_get(ml->call);
@@ -179,10 +213,8 @@ struct media_player *media_player_new(struct call_monologue *ml) {
 	mp->coder.pkt = av_packet_alloc();
 	mp->coder.pkt->data = NULL;
 	mp->coder.pkt->size = 0;
-
-	return mp;
 #else
-	return NULL;
+	return;
 #endif
 }
 
@@ -225,9 +257,21 @@ static void send_timer_rtcp(struct send_timer *st, struct ssrc_ctx *ssrc_out) {
 	timeval_add_usec(&ssrc_out->next_rtcp, 5000000 + (ssl_random() % 2000000));
 }
 
+struct async_send_req {
+	struct uring_req req; // must be first
+	struct iovec iov;
+	struct msghdr msg;
+	struct sockaddr_storage sin;
+	void *buf;
+};
+static void async_send_req_free(struct uring_req *p, int32_t res, uint32_t flags) {
+	struct async_send_req *req = (__typeof__(req)) p;
+	bufferpool_unref(req->buf);
+	uring_req_free(p);
+}
 
 static bool __send_timer_send_1(struct rtp_header *rh, struct packet_stream *sink, struct codec_packet *cp) {
-	struct stream_fd *sink_fd = sink->selected_sfd;
+	stream_fd *sink_fd = sink->selected_sfd;
 
 	if (!sink_fd || sink_fd->socket.fd == -1 || sink->endpoint.address.family == NULL)
 		return false;
@@ -249,11 +293,17 @@ static bool __send_timer_send_1(struct rtp_header *rh, struct packet_stream *sin
 				endpoint_print_buf(&sink_fd->socket.local),
 				FMT_M(endpoint_print_buf(&sink->endpoint)));
 
-	if (cp->kernel_send_info.local.family)
-		kernel_send_rtcp(&cp->kernel_send_info, cp->s.s, cp->s.len);
-	else
-		socket_sendto(&sink_fd->socket,
-				cp->s.s, cp->s.len, &sink->endpoint);
+	struct async_send_req *req = uring_alloc_req(sizeof(*req), async_send_req_free);
+	req->iov = (__typeof(req->iov)) {
+		.iov_base = cp->s.s,
+		.iov_len = cp->s.len,
+	};
+	req->msg = (__typeof(req->msg)) {
+		.msg_iov = &req->iov,
+		.msg_iovlen = 1,
+	};
+	req->buf = bufferpool_ref(cp->s.s);
+	uring_sendmsg(&sink_fd->socket, &req->msg, &sink->endpoint, &req->sin, &req->req);
 
 	if (sink->call->recording && rtpe_config.rec_egress) {
 		// fill in required members
@@ -266,10 +316,10 @@ static bool __send_timer_send_1(struct rtp_header *rh, struct packet_stream *sin
 		dump_packet(&mp, cp->plain.s ? &cp->plain : &cp->s);
 	}
 
-	atomic64_inc(&sink->stats_out.packets);
-	atomic64_add(&sink->stats_out.bytes, cp->s.len);
-	atomic64_inc(&sink_fd->local_intf->stats.out.packets);
-	atomic64_add(&sink_fd->local_intf->stats.out.bytes, cp->s.len);
+	atomic64_inc_na(&sink->stats_out->packets);
+	atomic64_add_na(&sink->stats_out->bytes, cp->s.len);
+	atomic64_inc_na(&sink_fd->local_intf->stats->out.packets);
+	atomic64_add_na(&sink_fd->local_intf->stats->out.bytes, cp->s.len);
 
 	log_info_pop();
 
@@ -281,12 +331,12 @@ static void __send_timer_send_common(struct send_timer *st, struct codec_packet
 		goto out;
 
 	if (cp->ssrc_out && cp->rtp) {
-		atomic64_inc(&cp->ssrc_out->packets);
-		atomic64_add(&cp->ssrc_out->octets, cp->s.len);
+		atomic64_inc_na(&cp->ssrc_out->stats->packets);
+		atomic64_add_na(&cp->ssrc_out->stats->bytes, cp->s.len);
 		if (cp->ts)
-			atomic64_set(&cp->ssrc_out->last_ts, cp->ts);
+			atomic_set_na(&cp->ssrc_out->stats->timestamp, cp->ts);
 		else
-			atomic64_set(&cp->ssrc_out->last_ts, ntohl(cp->rtp->timestamp));
+			atomic_set_na(&cp->ssrc_out->stats->timestamp, ntohl(cp->rtp->timestamp));
 		payload_tracker_add(&cp->ssrc_out->tracker, cp->rtp->m_pt & 0x7f);
 	}
 
@@ -305,7 +355,7 @@ out:
 }
 
 static void send_timer_send_lock(struct send_timer *st, struct codec_packet *cp) {
-	struct call *call = st->call;
+	call_t *call = st->call;
 	if (!call)
 		return;
 
@@ -322,7 +372,7 @@ static void send_timer_send_lock(struct send_timer *st, struct codec_packet *cp)
 }
 // st->stream->out_lock (or call->master_lock/W) must be held already
 static void send_timer_send_nolock(struct send_timer *st, struct codec_packet *cp) {
-	struct call *call = st->call;
+	call_t *call = st->call;
 	if (!call)
 		return;
 
@@ -408,15 +458,15 @@ retry:;
 
 		// EOF
 
-		if (mp->repeat <= 1) {
+		if (mp->opts.repeat <= 1) {
 			ilog(LOG_DEBUG, "EOF reading from media buffer (%s), stopping playback",
 					entry->info_str);
 			return true;
 		}
 
-		ilog(LOG_DEBUG, "EOF reading from media buffer (%s) but will repeat %li time",
-				entry->info_str, mp->repeat);
-		mp->repeat--;
+		ilog(LOG_DEBUG, "EOF reading from media buffer (%s) but will repeat %i time",
+				entry->info_str, mp->opts.repeat);
+		mp->opts.repeat--;
 		read_idx = mp->cache_read_idx = 0;
 		goto retry;
 	}
@@ -432,7 +482,7 @@ retry:;
 
 	// make a copy to send out
 	size_t len = pkt->s.len + sizeof(struct rtp_header) + RTP_BUFFER_TAIL_ROOM;
-	char *buf = g_malloc(len);
+	char *buf = bufferpool_alloc(media_bufferpool, len);
 	memcpy(buf, pkt->buf, len);
 
 	struct media_packet packet = {
@@ -466,10 +516,70 @@ retry:;
 	return false;
 }
 
-static void media_player_cached_reader_start(struct media_player *mp, const struct rtp_payload_type *dst_pt,
-		long long repeat)
-{
+static void media_player_kernel_player_start_now(struct media_player *mp) {
 	struct media_player_cache_entry *entry = mp->cache_entry;
+	if (!entry)
+		return;
+	const rtp_payload_type *dst_pt = &entry->coder.handler->dest_pt;
+
+	ilog(LOG_DEBUG, "Starting kernel media player index %i (PT %i)", entry->kernel_idx, dst_pt->payload_type);
+
+	struct rtpengine_play_stream_info info = {
+		.packet_stream_idx = entry->kernel_idx,
+		.pt = dst_pt->payload_type,
+		.seq = mp->seq,
+		.ts = mp->buffer_ts,
+		.ssrc = mp->ssrc_out->parent->h.ssrc,
+		.repeat = mp->opts.repeat,
+		.stats = mp->sink->stats_out,
+		.iface_stats = mp->sink->selected_sfd->local_intf->stats,
+		.ssrc_stats = mp->ssrc_out->stats,
+	};
+	mp->sink->endpoint.address.family->endpoint2kernel(&info.dst_addr, &mp->sink->endpoint); // XXX unify with __re_address_translate_ep
+	mp->sink->selected_sfd->socket.local.address.family->endpoint2kernel(&info.src_addr, &mp->sink->selected_sfd->socket.local); // XXX unify with __re_address_translate_ep
+	mp->crypt_handler->out->kernel(&info.encrypt, mp->sink);
+
+	unsigned int idx = kernel_start_stream_player(&info);
+	if (idx == -1)
+		ilog(LOG_ERR, "Failed to start kernel media player (index %i): %s", info.packet_stream_idx, strerror(errno));
+	else
+		mp->kernel_idx = idx;
+}
+
+static void media_player_kernel_player_start(struct media_player *mp) {
+	struct media_player_cache_entry *entry = mp->cache_entry;
+	if (!entry)
+		return;
+
+	// decoder finished yet? try unlocked read first
+	bool finished = entry->finished;
+
+	if (!finished) {
+		mutex_lock(&entry->lock);
+		// check flag again in case it happened just now
+		if (!entry->finished) {
+			// add us to wait list
+			ilog(LOG_DEBUG, "Decoder not finished yet, waiting to start kernel player index %i",
+					entry->kernel_idx);
+			t_hash_table_insert(entry->wait_queue, mp, mp); // XXX reference needed?
+			mutex_unlock(&entry->lock);
+			return;
+		}
+		// finished now, drop down below
+		mutex_unlock(&entry->lock);
+	}
+
+	media_player_kernel_player_start_now(mp);
+}
+
+static void media_player_cached_reader_start(struct media_player *mp) {
+	struct media_player_cache_entry *entry = mp->cache_entry;
+	const rtp_payload_type *dst_pt = &entry->coder.handler->dest_pt;
+
+	if (entry->kernel_idx != -1) {
+		media_player_kernel_player_start(mp);
+		return;
+	}
 
 	// create dummy codec handler and start timer
 
@@ -487,15 +597,13 @@ static void media_player_cached_reader_start(struct media_player *mp, const stru
 	}
 
 	mp->sync_ts_tv = rtpe_now;
-	mp->repeat = repeat;
 
 	media_player_read_decoded_packet(mp);
 }
 
 
-static void cache_packet_free(void *ptr) {
-	struct media_player_cache_packet *p = ptr;
-	g_free(p->buf);
+static void cache_packet_free(struct media_player_cache_packet *p) {
+	bufferpool_unref(p->buf);
 	g_slice_free1(sizeof(*p), p);
 }
 
@@ -503,12 +611,14 @@ static void cache_packet_free(void *ptr) {
 // returns: true = entry exists, decoding handled separately, use entry for playback
 //          false = no entry exists, OR entry is a new one, proceed to open decoder, then call _play_start
 static bool media_player_cache_get_entry(struct media_player *mp,
-		const struct rtp_payload_type *dst_pt, long long repeat)
+		const rtp_payload_type *dst_pt)
 {
 	if (!rtpe_config.player_cache)
 		return false;
 	if (mp->cache_index.type <= 0)
 		return false;
+	if (!dst_pt)
+		return false;
 
 	struct media_player_cache_index lookup;
 	lookup.index = mp->cache_index;
@@ -520,7 +630,7 @@ static bool media_player_cache_get_entry(struct media_player *mp,
 
 	bool ret = true; // entry exists, use cached data
 	if (entry) {
-		media_player_cached_reader_start(mp, dst_pt, repeat);
+		media_player_cached_reader_start(mp);
 		goto out;
 	}
 
@@ -532,10 +642,12 @@ static bool media_player_cache_get_entry(struct media_player *mp,
 	*ins_key = lookup;
 	str_init_dup_str(&ins_key->index.file, &lookup.index.file);
 	codec_init_payload_type(&ins_key->dst_pt, MT_UNKNOWN); // duplicate contents
+
 	entry = mp->cache_entry = g_slice_alloc0(sizeof(*entry));
 	mutex_init(&entry->lock);
 	cond_init(&entry->cond);
-	entry->packets = g_ptr_array_new_full(64, cache_packet_free);
+	entry->packets = cache_packet_arr_new_sized(64);
+	entry->wait_queue = media_player_ht_new();
 
 	switch (lookup.index.type) {
 		case MP_DB:
@@ -553,6 +665,15 @@ static bool media_player_cache_get_entry(struct media_player *mp,
 
 	g_hash_table_insert(media_player_cache, ins_key, entry);
 
+	entry->kernel_idx = -1;
+	if (kernel.use_player) {
+		entry->kernel_idx = kernel_get_packet_stream();
+		if (entry->kernel_idx == -1)
+			ilog(LOG_ERR, "Failed to get kernel packet stream entry (%s)", strerror(errno));
+		else
+			ilog(LOG_DEBUG, "Using kernel packet stream index %i", entry->kernel_idx);
+	}
+
 out:
 	mutex_unlock(&media_player_cache_lock);
 
@@ -600,12 +721,23 @@ static void media_player_cache_entry_decoder_thread(void *p) {
 		av_packet_unref(entry->coder.pkt);
 	}
 
+	ilog(LOG_DEBUG, "Decoder thread for %s finished", entry->info_str);
+
 	mutex_lock(&entry->lock);
 	entry->finished = true;
 	cond_broadcast(&entry->cond);
-	mutex_unlock(&entry->lock);
 
-	ilog(LOG_DEBUG, "Decoder thread for %s finished", entry->info_str);
+	media_player_ht_iter iter;
+	t_hash_table_iter_init(&iter, entry->wait_queue);
+	struct media_player *mp;
+	while (t_hash_table_iter_next(&iter, &mp, NULL)) {
+		if (mp->media)
+			media_player_kernel_player_start_now(mp);
+	}
+	t_hash_table_destroy(entry->wait_queue); // not needed any more
+	entry->wait_queue = media_player_ht_null();
+
+	mutex_unlock(&entry->lock);
 }
 
 static void packet_encoded_cache(AVPacket *pkt, struct codec_ssrc_handler *ch, struct media_packet *mp,
@@ -625,7 +757,17 @@ static void packet_encoded_cache(AVPacket *pkt, struct codec_ssrc_handler *ch, s
 	};
 
 	mutex_lock(&entry->lock);
-	g_ptr_array_add(entry->packets, ep);
+	t_ptr_array_add(entry->packets, ep);
+
+	if (entry->kernel_idx != -1) {
+		ilog(LOG_DEBUG, "Adding media packet (length %zu, TS %" PRIu64 ", delay %lu ms) to kernel packet stream %i",
+				s->len, pkt->pts, entry->duration, entry->kernel_idx);
+		if (!kernel_add_stream_packet(entry->kernel_idx, s->s, s->len, entry->duration, pkt->pts,
+					pkt->duration))
+			ilog(LOG_ERR, "Failed to add packet to kernel player (%s)", strerror(errno));
+	}
+
+	entry->duration += ep->duration / 1000;
 
 	cond_broadcast(&entry->cond);
 	mutex_unlock(&entry->lock);
@@ -647,9 +789,7 @@ static int media_player_packet_cache(encoder_t *enc, void *u1, void *u2) {
 // do have a cache entry, initialise it, set up the thread, take over decoding, and then proceed as a
 // media player consuming the data from the decoder thread.
 // returns: false = continue normally decode in-thread, true = take data from other thread
-static bool media_player_cache_entry_init(struct media_player *mp, const struct rtp_payload_type *dst_pt,
-		long long repeat)
-{
+static bool media_player_cache_entry_init(struct media_player *mp, const rtp_payload_type *dst_pt) {
 	struct media_player_cache_entry *entry = mp->cache_entry;
 	if (!entry)
 		return false;
@@ -664,7 +804,7 @@ static bool media_player_cache_entry_init(struct media_player *mp, const struct
 	// use low priority (10 nice)
 	thread_create_detach_prio(media_player_cache_entry_decoder_thread, entry, NULL, 10, "mp decoder");
 
-	media_player_cached_reader_start(mp, dst_pt, repeat);
+	media_player_cached_reader_start(mp);
 
 	return true;
 }
@@ -672,9 +812,9 @@ static bool media_player_cache_entry_init(struct media_player *mp, const struct
 
 
 // find suitable output payload type
-static struct rtp_payload_type *media_player_get_dst_pt(struct media_player *mp) {
-	struct rtp_payload_type *dst_pt = NULL;
-	for (GList *l = mp->media->codecs.codec_prefs.head; l; l = l->next) {
+static rtp_payload_type *media_player_get_dst_pt(struct media_player *mp) {
+	rtp_payload_type *dst_pt = NULL;
+	for (__auto_type l = mp->media->codecs.codec_prefs.head; l; l = l->next) {
 		dst_pt = l->data;
 		ensure_codec_def(dst_pt, mp->media);
 		if (dst_pt->codec_def && !dst_pt->codec_def->supplemental)
@@ -691,8 +831,8 @@ found:
 }
 
 
-bool media_player_pt_match(const struct media_player *mp, const struct rtp_payload_type *src_pt,
-		const struct rtp_payload_type *dst_pt)
+bool media_player_pt_match(const struct media_player *mp, const rtp_payload_type *src_pt,
+		const rtp_payload_type *dst_pt)
 {
 	if (!mp->coder.handler)
 		return true; // not initialised yet -> doesn't need a reset
@@ -704,8 +844,8 @@ bool media_player_pt_match(const struct media_player *mp, const struct rtp_paylo
 }
 
 
-static int media_player_setup_common(struct media_player *mp, const struct rtp_payload_type *src_pt,
-		const struct rtp_payload_type **dst_pt)
+static int media_player_setup_common(struct media_player *mp, const rtp_payload_type *src_pt,
+		const rtp_payload_type **dst_pt)
 {
 	if (!*dst_pt)
 		*dst_pt = media_player_get_dst_pt(mp);
@@ -729,8 +869,8 @@ static int media_player_setup_common(struct media_player *mp, const struct rtp_p
 }
 
 // used for generic playback (audio_player, t38_gateway)
-int media_player_setup(struct media_player *mp, const struct rtp_payload_type *src_pt,
-		const struct rtp_payload_type *dst_pt)
+int media_player_setup(struct media_player *mp, const rtp_payload_type *src_pt,
+		const rtp_payload_type *dst_pt)
 {
 	int ret = media_player_setup_common(mp, src_pt, &dst_pt);
 	if (ret)
@@ -745,8 +885,8 @@ int media_player_setup(struct media_player *mp, const struct rtp_payload_type *s
 	return 0;
 }
 // used for "play media" player
-static int __media_player_setup_internal(struct media_player *mp, const struct rtp_payload_type *src_pt,
-		const struct rtp_payload_type *dst_pt)
+static int __media_player_setup_internal(struct media_player *mp, const rtp_payload_type *src_pt,
+		const rtp_payload_type *dst_pt)
 {
 	int ret = media_player_setup_common(mp, src_pt, &dst_pt);
 	if (ret)
@@ -761,12 +901,12 @@ static int __media_player_setup_internal(struct media_player *mp, const struct r
 	return 0;
 }
 
-static int __ensure_codec_handler(struct media_player *mp, const struct rtp_payload_type *dst_pt) {
+static int __ensure_codec_handler(struct media_player *mp, const rtp_payload_type *dst_pt) {
 	if (mp->coder.handler)
 		return 0;
 
 	// synthesise rtp payload type
-	struct rtp_payload_type src_pt = { .payload_type = -1 };
+	rtp_payload_type src_pt = { .payload_type = -1 };
 	src_pt.codec_def = codec_find_by_av(mp->coder.avstream->CODECPAR->codec_id);
 	if (!src_pt.codec_def) {
 		ilog(LOG_ERR, "Attempting to play media from an unsupported file format/codec");
@@ -845,9 +985,10 @@ static bool media_player_read_packet(struct media_player *mp) {
 	int ret = av_read_frame(mp->coder.fmtctx, mp->coder.pkt);
 	if (ret < 0) {
 		if (ret == AVERROR_EOF) {
-			if (mp->repeat > 1){
-				ilog(LOG_DEBUG, "EOF reading from media stream but will repeat %li time",mp->repeat);
-				mp->repeat = mp->repeat - 1;
+			if (mp->opts.repeat > 1) {
+				ilog(LOG_DEBUG, "EOF reading from media stream but will repeat %i time",
+						mp->opts.repeat);
+				mp->opts.repeat = mp->opts.repeat - 1;
 				int64_t ret64 = avio_seek(mp->coder.fmtctx->pb, 0, SEEK_SET);
 				if (ret64 != 0)
 					ilog(LOG_ERR, "Failed to seek to beginning of media file");
@@ -891,9 +1032,7 @@ void media_player_set_media(struct media_player *mp, struct call_media *media) {
 
 // call->master_lock held in W
 // returns destination payload type, or NULL on failure
-static const struct rtp_payload_type *media_player_play_init(struct media_player *mp) {
-	media_player_shutdown(mp);
-
+static const rtp_payload_type *media_player_play_setup(struct media_player *mp) {
 	// find call media suitable for playback
 	struct call_media *media;
 	for (unsigned int i = 0; i < mp->ml->medias->len; i++) {
@@ -916,52 +1055,67 @@ found:
 	return media_player_get_dst_pt(mp);
 }
 
+// call->master_lock held in W
+// returns destination payload type, or NULL on failure
+static const rtp_payload_type *media_player_play_init(struct media_player *mp) {
+	media_player_shutdown(mp);
+	return media_player_play_setup(mp);
+}
+
 
 // call->master_lock held in W
-static void media_player_play_start(struct media_player *mp, const struct rtp_payload_type *dst_pt,
-		long long repeat, long long start_pos)
-{
+static bool media_player_play_start(struct media_player *mp, const rtp_payload_type *dst_pt) {
 	// needed to have usable duration for some formats. ignore errors.
-	avformat_find_stream_info(mp->coder.fmtctx, NULL);
+	if (!mp->coder.fmtctx->streams || !mp->coder.fmtctx->streams[0])
+		avformat_find_stream_info(mp->coder.fmtctx, NULL);
 
 	mp->coder.avstream = mp->coder.fmtctx->streams[0];
 	if (!mp->coder.avstream) {
 		ilog(LOG_ERR, "No AVStream present in format context");
-		return;
+		return false;
 	}
-	if (__ensure_codec_handler(mp, dst_pt))
-		return;
 
-	if (media_player_cache_entry_init(mp, dst_pt, repeat))
-		return;
+	if (__ensure_codec_handler(mp, dst_pt))
+		return false;
+
+	if (mp->opts.block_egress)
+		MEDIA_SET(mp->media, BLOCK_EGRESS);
+
+	if (media_player_cache_entry_init(mp, dst_pt))
+		return true;
 
 	mp->next_run = rtpe_now;
 	// give ourselves a bit of a head start with decoding
 	timeval_add_usec(&mp->next_run, -50000);
 
 	// if start_pos is positive, try to seek to that position
-	if (start_pos > 0) {
-		ilog(LOG_DEBUG, "Seeking to position %lli", start_pos);
-		av_seek_frame(mp->coder.fmtctx, 0, start_pos, 0);
+	if (mp->opts.start_pos > 0) {
+		ilog(LOG_DEBUG, "Seeking to position %lli", mp->opts.start_pos);
+		av_seek_frame(mp->coder.fmtctx, 0, mp->opts.start_pos, 0);
 	}
+	else // in case this is a repeated start
+		av_seek_frame(mp->coder.fmtctx, 0, 0, 0);
+
 	media_player_read_packet(mp);
-	mp->repeat = repeat;
+
+	return true;
 }
 #endif
 
 
 // call->master_lock held in W
-int media_player_play_file(struct media_player *mp, const str *file, long long repeat, long long start_pos) {
+static mp_cached_code __media_player_init_file(struct media_player *mp, const str *file,
+		media_player_opts_t opts,
+		const rtp_payload_type *dst_pt)
+{
 #ifdef WITH_TRANSCODING
-	const struct rtp_payload_type *dst_pt = media_player_play_init(mp);
-	if (!dst_pt)
-		return -1;
-
 	mp->cache_index.type = MP_FILE;
 	str_init_dup_str(&mp->cache_index.file, file);
 
-	if (media_player_cache_get_entry(mp, dst_pt, repeat))
-		return 0;
+	mp->opts = opts;
+
+	if (media_player_cache_get_entry(mp, dst_pt))
+		return MPC_CACHED;
 
 	char file_s[PATH_MAX];
 	snprintf(file_s, sizeof(file_s), STR_FORMAT, STR_FMT(file));
@@ -969,17 +1123,40 @@ int media_player_play_file(struct media_player *mp, const str *file, long long r
 	int ret = avformat_open_input(&mp->coder.fmtctx, file_s, NULL, NULL);
 	if (ret < 0) {
 		ilog(LOG_ERR, "Failed to open media file for playback: %s", av_error(ret));
-		return -1;
+		return MPC_ERR;
 	}
 
-	media_player_play_start(mp, dst_pt, repeat, start_pos);
-
-	return 0;
+	return MPC_OK;
 #else
-	return -1;
+	return MPC_ERR;
 #endif
 }
 
+// call->master_lock held in W
+bool media_player_play_file(struct media_player *mp, const str *file, media_player_opts_t opts) {
+#ifdef WITH_TRANSCODING
+	const rtp_payload_type *dst_pt = media_player_play_init(mp);
+	if (!dst_pt)
+		return false;
+
+	mp_cached_code ret = __media_player_init_file(mp, file, opts, dst_pt);
+	if (ret == MPC_CACHED)
+		return true;
+	if (ret == MPC_ERR)
+		return false;
+
+	return media_player_play_start(mp, dst_pt);
+#else
+	return false;
+#endif
+}
+
+// call->master_lock held in W
+bool media_player_init_file(struct media_player *mp, const str *file, media_player_opts_t opts) {
+	int ret = __media_player_init_file(mp, file, opts, NULL);
+	return ret == 0;
+}
+
 
 #ifdef WITH_TRANSCODING
 static int __mp_avio_read_wrap(void *opaque, uint8_t *buf, int buf_size) {
@@ -1029,29 +1206,28 @@ static int64_t __mp_avio_seek(void *opaque, int64_t offset, int whence) {
 
 
 // call->master_lock held in W
-static int media_player_play_blob_id(struct media_player *mp, const str *blob, long long repeat,
-		long long start_pos, long long db_id)
+static mp_cached_code __media_player_init_blob_id(struct media_player *mp, const str *blob,
+		media_player_opts_t opts,
+		long long db_id, const rtp_payload_type *dst_pt)
 {
 	const char *err;
 	int av_ret = 0;
 
-	const struct rtp_payload_type *dst_pt = media_player_play_init(mp);
-	if (!dst_pt)
-		return -1;
+	mp->opts = opts;
 
 	if (db_id >= 0) {
 		mp->cache_index.type = MP_DB;
 		mp->cache_index.db_id = db_id;
 
-		if (media_player_cache_get_entry(mp, dst_pt, repeat))
-			return 0;
+		if (media_player_cache_get_entry(mp, dst_pt))
+			return MPC_CACHED;
 	}
 	else {
 		mp->cache_index.type = MP_BLOB;
 		str_init_dup_str(&mp->cache_index.file, blob);
 
-		if (media_player_cache_get_entry(mp, dst_pt, repeat))
-			return 0;
+		if (media_player_cache_get_entry(mp, dst_pt))
+			return MPC_CACHED;
 	}
 
 	mp->coder.blob = str_dup(blob);
@@ -1084,30 +1260,38 @@ static int media_player_play_blob_id(struct media_player *mp, const str *blob, l
 	if (av_ret < 0)
 		goto err;
 
-	media_player_play_start(mp, dst_pt, repeat, start_pos);
-
-	return 0;
+	return MPC_OK;
 
 err:
 	ilog(LOG_ERR, "Failed to start media playback from memory: %s", err);
 	if (av_ret)
 		ilog(LOG_ERR, "Error returned from libav: %s", av_error(av_ret));
-	return -1;
+	return MPC_ERR;
 }
-#endif
 
 
 // call->master_lock held in W
-int media_player_play_blob(struct media_player *mp, const str *blob, long long repeat, long long start_pos) {
-#ifdef WITH_TRANSCODING
-	return media_player_play_blob_id(mp, blob, repeat, start_pos, -1);
-#else
-	return -1;
-#endif
+bool media_player_play_blob(struct media_player *mp, const str *blob, media_player_opts_t opts) {
+	const rtp_payload_type *dst_pt = media_player_play_init(mp);
+	if (!dst_pt)
+		return false;
+
+	mp_cached_code ret = __media_player_init_blob_id(mp, blob, opts, -1, dst_pt);
+	if (ret == MPC_CACHED)
+		return true;
+	if (ret == MPC_ERR)
+		return false;
+
+	return media_player_play_start(mp, dst_pt);
+}
+
+// call->master_lock held in W
+bool media_player_init_blob(struct media_player *mp, const str *blob, media_player_opts_t opts) {
+	int ret = __media_player_init_blob_id(mp, blob, opts, -1, NULL);
+	return ret == 0;
 }
 
 
-#ifdef WITH_TRANSCODING
 static int __connect_db(void) {
 	if (mysql_conn) {
 		mysql_close(mysql_conn);
@@ -1131,9 +1315,11 @@ err:
 
 
 // call->master_lock held in W
-int media_player_play_db(struct media_player *mp, long long id, long long repeat, long long start_pos) {
+static mp_cached_code __media_player_init_db(struct media_player *mp, long long id, media_player_opts_t opts,
+		const rtp_payload_type *dst_pt)
+{
 	const char *err;
-	AUTO_CLEANUP_GBUF(query);
+	g_autoptr(char) query = NULL;
 
 	err = "missing configuration";
 	if (!rtpe_config.mysql_host || !rtpe_config.mysql_query)
@@ -1177,7 +1363,7 @@ success:;
 	}
 
 	str blob = STR_INIT_LEN(row[0], lengths[0]);
-	int ret = media_player_play_blob_id(mp, &blob, repeat, start_pos, id);
+	mp_cached_code ret = __media_player_init_blob_id(mp, &blob, opts, id, dst_pt);
 
 	mysql_free_result(res);
 
@@ -1188,13 +1374,34 @@ err:
 		ilog(LOG_ERR, "Failed to start media playback from database (used query '%s'): %s", query, err);
 	else
 		ilog(LOG_ERR, "Failed to start media playback from database: %s", err);
-	return -1;
+	return MPC_ERR;
+}
+
+// call->master_lock held in W
+bool media_player_play_db(struct media_player *mp, long long id, media_player_opts_t opts) {
+	const rtp_payload_type *dst_pt = media_player_play_init(mp);
+	if (!dst_pt)
+		return false;
+
+	mp_cached_code ret = __media_player_init_db(mp, id, opts, dst_pt);
+	if (ret == MPC_CACHED)
+		return true;
+	if (ret == MPC_ERR)
+		return false;
+
+	return media_player_play_start(mp, dst_pt);
+}
+
+// call->master_lock held in W
+bool media_player_init_db(struct media_player *mp, long long id, media_player_opts_t opts) {
+	int ret = __media_player_init_db(mp, id, opts, NULL);
+	return ret == 0;
 }
 
 
 static void media_player_run(void *ptr) {
 	struct media_player *mp = ptr;
-	struct call *call = mp->call;
+	call_t *call = mp->call;
 
 	log_info_media(mp->media);
 
@@ -1215,8 +1422,11 @@ static void media_player_run(void *ptr) {
 
 		mp->next_run.tv_sec = 0;
 
+		if (mp->opts.block_egress)
+			MEDIA_CLEAR(mp->media, BLOCK_EGRESS);
+
 		codec_update_all_source_handlers(mp->media->monologue, NULL);
-		update_init_subscribers(mp->media->monologue, OP_OTHER);
+		update_init_subscribers(mp->media->monologue, OP_PLAY_MEDIA);
 
 		rwlock_unlock_w(&call->master_lock);
 	}
@@ -1280,16 +1490,35 @@ static void media_player_cache_index_free(void *p) {
 }
 static void media_player_cache_entry_free(void *p) {
 	struct media_player_cache_entry *e = p;
-	g_ptr_array_free(e->packets, TRUE);
+	t_ptr_array_free(e->packets, true);
 	mutex_destroy(&e->lock);
 	g_free(e->info_str);
+	if (t_hash_table_is_set(e->wait_queue))
+		t_hash_table_destroy(e->wait_queue); // XXX release references?
 	media_player_coder_shutdown(&e->coder);
 	av_packet_free(&e->coder.pkt);
+	kernel_free_packet_stream(e->kernel_idx);
 	g_slice_free1(sizeof(*e), e);
 }
 #endif
 
 
+// call->master_lock held in W
+bool media_player_start(struct media_player *mp) {
+#ifdef WITH_TRANSCODING
+	if (!mp->coder.fmtctx) // initialised?
+		return false;
+
+	const rtp_payload_type *dst_pt = media_player_play_setup(mp);
+	if (!dst_pt)
+		return false;
+
+	return media_player_play_start(mp, dst_pt);
+#else
+	return false;
+#endif
+}
+
 void media_player_init(void) {
 #ifdef WITH_TRANSCODING
 	if (rtpe_config.player_cache) {
@@ -1299,9 +1528,9 @@ void media_player_init(void) {
 		mutex_init(&media_player_cache_lock);
 	}
 
-	timerthread_init(&media_player_thread, media_player_run);
+	timerthread_init(&media_player_thread, rtpe_config.media_num_threads, media_player_run);
 #endif
-	timerthread_init(&send_timer_thread, timerthread_queue_run);
+	timerthread_init(&send_timer_thread, rtpe_config.media_num_threads, timerthread_queue_run);
 }
 
 void media_player_free(void) {
@@ -1317,12 +1546,12 @@ void media_player_free(void) {
 }
 
 
+void media_player_launch(void) {
 #ifdef WITH_TRANSCODING
-void media_player_loop(void *p) {
-	timerthread_run(&media_player_thread);
-}
+	timerthread_launch(&media_player_thread, rtpe_config.scheduling, rtpe_config.priority, "media player");
 #endif
-void send_timer_loop(void *p) {
+}
+void send_timer_launch(void) {
 	//ilog(LOG_DEBUG, "send_timer_loop");
-	timerthread_run(&send_timer_thread);
+	timerthread_launch(&send_timer_thread, rtpe_config.scheduling, rtpe_config.priority, "media player");
 }
