@@ -8,6 +8,7 @@
 #include <assert.h>
 #include <stdio.h>
 #include "compat.h"
+#include "containers.h"
 
 
 
@@ -18,6 +19,8 @@ struct _str {
 
 typedef struct _str str;
 
+TYPED_GQUEUE(str, str)
+
 
 
 #define STR_FORMAT "%.*s"
@@ -34,8 +37,9 @@ typedef struct _str str;
 #define STR_NULL ((str) { NULL, 0 })
 #define STR_EMPTY ((str) { "", 0 })
 #define STR_CONST_INIT(s) ((str) { s, sizeof(s)-1 })
-#define STR_INIT(s) ((str) { s, (s) ? strlen(s) : 0 })
-#define STR_INIT_LEN(s, len) ((str) { s, len })
+#define STR_INIT(s) ((str) { (char *) (s), (s) ? strlen(s) : 0 })
+#define STR_INIT_GS(s) ((str) { (s)->str, (s)->len })
+#define STR_INIT_LEN(s, len) ((str) { (char *) (s), len })
 #define STR_INIT_DUP(s) ((str) { g_strdup(s), strlen(s) })
 #define STR_CONST_INIT_BUF(buf) ((str) { (char *) &buf, sizeof(buf) })
 
@@ -104,9 +108,12 @@ INLINE str *str_alloc(size_t len);
 __attribute__((nonnull(1)))
 ACCESS(read_only, 1)
 INLINE str *str_dup(const str *s);
-/* shifts pointer by len chars and decrements len. returns -1 if buffer too short, 0 otherwise */
+/* free function corresponding to str_dup() */
 __attribute__((nonnull(1)))
 ACCESS(read_write, 1)
+INLINE void str_free(str *s);
+/* shifts pointer by len chars and decrements len. returns -1 if buffer too short, 0 otherwise */
+ACCESS(read_write, 1)
 INLINE int str_shift(str *s, size_t len);
 /* to revert a previously successful str_shift(). no error checking */
 __attribute__((nonnull(1)))
@@ -149,12 +156,12 @@ INLINE unsigned long long str_to_ui(const str *s, unsigned long long def);
 __attribute__((nonnull(1, 2)))
 ACCESS(write_only, 1)
 ACCESS(read_write, 2)
-INLINE int str_token(str *new_token, str *ori_and_remainder, int sep);
+INLINE bool str_token(str *new_token, str *ori_and_remainder, int sep);
 /* same as str_token but allows for a trailing non-empty token (e.g. "foo,bar" -> "foo", "bar" ) */
 __attribute__((nonnull(1, 2)))
 ACCESS(write_only, 1)
 ACCESS(read_write, 2)
-INLINE int str_token_sep(str *new_token, str *ori_and_remainder, int sep);
+INLINE bool str_token_sep(str *new_token, str *ori_and_remainder, int sep);
 /* copy a string to a regular C string buffer, limiting the max size */
 __attribute__((nonnull(1, 3)))
 ACCESS(write_only, 1, 2)
@@ -171,10 +178,14 @@ INLINE GString *g_string_new_str(void);
 INLINE str *g_string_free_str(GString *gs);
 
 /* for GHashTables */
-guint str_hash(gconstpointer s);
-gboolean str_equal(gconstpointer a, gconstpointer b);
-guint str_case_hash(gconstpointer s);
-gboolean str_case_equal(gconstpointer a, gconstpointer b);
+guint str_hash(const str *s);
+gboolean str_equal(const str *a, const str *b);
+guint str_case_hash(const str *s);
+gboolean str_case_equal(const str *a, const str *b);
+
+TYPED_GHASHTABLE(str_case_ht, str, str, str_case_hash, str_case_equal, free, NULL)
+TYPED_GHASHTABLE(str_case_value_ht, str, str, str_case_hash, str_case_equal, free, free)
+
 
 /* returns a new str object, duplicates the pointers but doesn't duplicate the contents */
 INLINE str *str_slice_dup(const str *);
@@ -188,6 +199,8 @@ INLINE size_t str_uri_encode(char *out, const str *in);
 str *str_uri_decode_len(const char *in, size_t in_len);
 
 
+G_DEFINE_AUTO_CLEANUP_CLEAR_FUNC(str, str_free_dup);
+
 
 
 
@@ -344,6 +357,9 @@ INLINE str *str_dup(const str *s) {
 	r->s[s->len] = '\0';
 	return r;
 }
+INLINE void str_free(str *s) {
+	free(s);
+}
 INLINE str *str_slice_dup(const str *s) {
 	str *r;
 	r = g_slice_alloc(sizeof(*r));
@@ -440,25 +456,25 @@ INLINE unsigned long long str_to_ui(const str *s, unsigned long long def) {
 	return ret;
 }
 
-INLINE int str_token(str *new_token, str *ori_and_remainder, int sep) {
+INLINE bool str_token(str *new_token, str *ori_and_remainder, int sep) {
 	*new_token = *ori_and_remainder;
 	if (!str_chr_str(ori_and_remainder, ori_and_remainder, sep))
-		return -1;
+		return false;
 	new_token->len = ori_and_remainder->s - new_token->s;
 	if (str_shift(ori_and_remainder, 1))
-		return -1;
-	return 0;
+		return false;
+	return true;
 }
 
-INLINE int str_token_sep(str *new_token, str *ori_and_remainder, int sep) {
+INLINE bool str_token_sep(str *new_token, str *ori_and_remainder, int sep) {
 	str ori = *ori_and_remainder;
-	if (!str_token(new_token, ori_and_remainder, sep))
-		return 0;
+	if (str_token(new_token, ori_and_remainder, sep))
+		return true;
 	// separator not found, use remainder as final token if not empty
 	if (!ori.len)
-		return -1;
+		return false;
 	*new_token = ori;
-	return 0;
+	return true;
 }
 
 INLINE size_t str_uri_encode(char *out, const str *in) {
